unit Helios_ROQ;
{
 * ------------------------------------------------------------------------
 * Title: RoQ Video File Delphi Conversion
 * Ver:   v1.0b
 * By:    Stuart Carey (Stucuk)
 *
 * Convertion Start: 18/09/04
 * Comments:
 *
 *
 * ------------------------------------------------------------------------
 * Id Software's RoQ video file format decoder
 *
 * Dr. Tim Ferguson, 2001.
 * For more details on the algorithm:
 *         http://www.csse.monash.edu.au/~timf/videocodec.html
 *
 * This is a simple decoder for the Id Software RoQ video format.  In
 * this format, audio samples are DPCM coded and the video frames are
 * coded using motion blocks and vector quantisation.
 *
 * Note: All information on the RoQ file format has been obtained through
 *   pure reverse engineering.  This was achieved by giving known input
 *   audio and video frames to the roq.exe encoder and analysing the
 *   resulting output text and RoQ file.  No decompiling of the Quake III
 *   Arena game was required.
 *
 * You may freely use this source code.  I only ask that you reference its
 * source in your projects documentation:
 *       Tim Ferguson: http://www.csse.monash.edu.au/~timf/
 * ------------------------------------------------------------------------
}


interface

const
RoQ_INFO          = $1001;
RoQ_QUAD_CODEBOOK = $1002;
RoQ_QUAD_VQ       = $1011;
RoQ_SOUND_MONO    = $1020;
RoQ_SOUND_STEREO  = $1021;

RoQ_ID_MOT = $00;
RoQ_ID_FCC = $01;
RoQ_ID_SLD = $02;
RoQ_ID_CCC = $03;

//New
SEEK_CUR = 1;
SEEK_SET = 2;

type
Proq_cell = ^roq_cell;
roq_cell = record
 y0,y1,y2,y3,u,v : Byte;
end;

type
Proq_qcell = ^roq_qcell;
roq_qcell = record
idx : array [0..3] of integer;
end;

Type
PFile = ^File;

type
Proq_info = ^Troq_info;
Troq_info = record
  FP : File;
  Buf_Size : Integer;
  Buf : array of Byte;
  cells : array [0..255] of roq_cell;
  qcells : array [0..255] of roq_qcell;
  snd_sqr_arr : array [0..259] of Shortint;
  roq_start, aud_pos, vid_pos : Longint;
  Frame_offset : array of Longint;
  num_frames,num_audio_bytes : longword;
  width, height, frame_num, audio_channels : integer;
  y,u,v : array [0..1] of array of Byte;
  stream_length : longint;
  audio_buf_size, audio_size : integer;
  audio : array of Byte;
end;

Function roq_open(fname : pchar) : Proq_info;
Procedure roq_close(var RI : Proq_info);
Function roq_read_audio(var Ri : Proq_info) : integer;

implementation

uses Windows, Sysutils;

// Delphi's Equivalent to C Procedures/Functions

// New
Function FGetC(FP : PFile) : Byte;
begin
BlockRead(FP^,Result,Sizeof(Byte)); // Read Byte
end;

//New
Function FTell(FP : PFile) : Longint;
begin
Result := FilePos(FP^);
end;

//New
Function Feof(FP : PFile) : Boolean;
begin
Result := Eof(FP^);
end;

//New
Function Fseek(FP : PFile; Offset : Longint; whence : integer) : Boolean;
var
Pos : Longint;
begin

Pos := Offset;
if whence = SEEK_CUR then
Pos := Pos + FTell(FP);

Seek(FP^,Pos);
end;

// Start of roq_read.h
Function Get_Word(FP : PFile) : Word;
begin
Result := (FGetC(FP) and $FF);
Result := Result or (FGetC(FP) and $FF) shl 8;
end;

Function Get_Long(FP : PFile) : Longword;
begin
Result := (FGetC(FP) and $FF);
Result := Result or (FGetC(FP) and $FF) shl 8;
Result := Result or (FGetC(FP) and $FF) shl 16;
Result := Result or (FGetC(FP) and $FF) shl 24;
end;

Function roq_parse_file(var RI : Proq_info) : integer;
var
chunk_id,chunk_arg : Cardinal;
chunk_size, fpos : longint;
head1,head3 : Word;
head2 : LongWord;
Max_Frames : integer;
fp : PFile;
begin
RI^.num_frames := 0;
RI^.num_audio_bytes := 0;
RI^.audio_channels := 0;
SetLength(RI^.Frame_offset,0);
Max_Frames := 0;
RI^.Buf_Size := 0;

fp := @RI^.fp;

head1 := get_word(fp);
head2 := get_long(fp);
head3 := get_word(fp);

if(not (head1 = $1084) and not (head2 = $ffffffff) and not (head3 = $1e)) then
begin
MessageBox(0,'Not an RoQ file','Invalid Header',0);
Result := 1;
Exit;
end;

RI^.roq_start := FTell(fp);

while not Feof(fp) do
begin

try
fpos       := FTell(fp);
chunk_id   := get_word(fp);
chunk_size := get_long(fp);
chunk_arg  := get_word(fp);
except
exit;
end;

if(chunk_size > ri^.buf_size) then ri^.buf_size := chunk_size;

if(feof(fp)) then break;

if(chunk_id = RoQ_INFO) then //	/* video info */
begin
ri^.width  := get_word(fp);
ri^.height := get_word(fp);
get_word(fp);
get_word(fp);
end
else
begin
if(chunk_id = RoQ_QUAD_VQ) then
begin
inc(ri^.num_frames);

if ri^.num_frames > Max_Frames then
begin
Max_Frames := Max_Frames + 5000;
SetLength(ri^.frame_offset,Max_Frames);
end;

ri^.frame_offset[ri^.num_frames-1] := fpos;

end
else
if((chunk_id = RoQ_SOUND_MONO) or (chunk_id = RoQ_SOUND_STEREO)) then
begin
if(chunk_id = RoQ_SOUND_MONO) then ri^.audio_channels := 1 else ri^.audio_channels := 2;
ri^.num_audio_bytes := ri^.num_audio_bytes + chunk_size;
end;
fseek(fp, chunk_size, SEEK_CUR);
end;

end;

Result := 0;
SetLength(ri^.frame_offset,ri^.num_frames); // no point wasting memory!

end;

procedure apply_vector_2x2(var ri:Proq_info; x, y: integer; var cell:Proq_cell);
var
   p : integer;
begin
   p := (y * ri^.width) + x;
   ri^.y[0][p] := cell^.y0;
   inc(p);
   ri^.y[0][p] := cell^.y1;
   inc(p);
   p := p + (ri^.width - 2);
   ri^.y[0][p] := cell^.y2;
   inc(p);
   ri^.y[0][p] := cell^.y3;
   inc(p);
   ri^.u[0][(y div 2) * (ri^.width div 2) + (x div 2)] := cell^.u;
   ri^.v[0][(y div 2) * (ri^.width div 2) + (x div 2)] := cell^.v;
end;

procedure apply_vector_4x4(var ri:Proq_info; x, y: integer; var cell:Proq_cell);
var
p,p2,p3 : integer;
row_inc, c_row_inc : longint;
y0, y1, u, v : byte;
begin

	P  := (y * ri^.width) + x;
	P2 := (y div 2) * (ri^.width div 2) + x div 2;
	P3 := (y div 2) * (ri^.width div 2) + x div 2;

	row_inc   := ri^.width - 4;
	c_row_inc := (ri^.width div 2) - 2;

        y0 := cell^.y0;
        y1 := cell^.y1;
        v := cell^.v;
        u := cell^.u;

        ri^.y[0][p] := y0;
        Inc(p);
        ri^.y[0][p] := y0;
        Inc(p);

        ri^.y[0][p] := y1;
        Inc(p);
        ri^.y[0][p] := y1;
        Inc(p);

        p := p + row_inc;

        ri^.y[0][p] := y0;
        Inc(p);
        ri^.y[0][p] := y0;
        Inc(p);
        ri^.y[0][p] := y1;
        Inc(p);
        ri^.y[0][p] := y1;
        Inc(p);

        p := p + row_inc;

        y0 := cell^.y2;
        y1 := cell^.y3;
        ri^.y[0][p] := y0;
        Inc(p);
        ri^.y[0][p] := y0;
        Inc(p);
        ri^.y[0][p] := y1;
        Inc(p);
        ri^.y[0][p] := y1;
        Inc(p);

        p := p + row_inc;

        ri^.y[0][p] := y0;
        Inc(p);
        ri^.y[0][p] := y0;
        Inc(p);
        ri^.y[0][p] := y1;
        Inc(p);
        ri^.y[0][p] := y1;
        Inc(p);

        u := cell^.u;
        v := cell^.v;

        ri^.u[0][p2] := u;
        Inc(p2);
        ri^.u[0][p2] := u;
        Inc(p2);

        p2 := p2 + c_row_inc;

        ri^.u[0][p2] := u;
        Inc(p2);
        ri^.u[0][p2] := u;
        Inc(p2);

        ri^.v[0][p3] := v;
        Inc(p3);
        ri^.v[0][p3] := v;
        Inc(p3);

        p3 := p3 + c_row_inc;

        ri^.v[0][p3] := v;
        Inc(p3);
        ri^.v[0][p3] := v;
        Inc(p3);
end;

procedure apply_motion_4x4(var ri:Proq_info; x, y: integer; mv,mean_x,mean_y : byte);
var
i,mx,my : integer;
pa,pb : integer;
begin
        mx := x + 8 - (mv shr 4) - mean_x;
	my := y + 8 - (mv and $f) - mean_y;

        pa  := (y * ri^.width) + x; //ri^.y[0]
        pb := (my * ri^.width) + mx; //ri^.y[1]
	for i := 0 to 3 do
	begin
         ri^.y[0][pa+0] := ri^.y[1][pb+0];
         ri^.y[0][pa+1] := ri^.y[1][pb+1];
         ri^.y[0][pa+2] := ri^.y[1][pb+2];
         ri^.y[0][pa+3] := ri^.y[1][pb+3];
         pa := pa + ri^.width;
	 pb := pb + ri^.width;
        end;

        pa := (y div 2) * (ri^.width div 2) + x div 2;
	pb := (my div 2) * (ri^.width div 2) + (mx + 1) div 2;
	for i := 0 to 1 do
	begin
	 ri^.u[0][pa+0] := ri^.u[1][pb+0];
         ri^.u[0][pa+1] := ri^.u[1][pb+1];
	 pa := pa + ri^.width div 2;
	 pb := pb + ri^.width div 2;
        end;

        pa := (y div 2) * (ri^.width div 2) + x div 2;
	pb := (my div 2) * (ri^.width div 2) + (mx + 1) div 2;
	for i := 0 to 1 do
	begin
	 ri^.v[0][pa+0] := ri^.v[1][pb+0];
         ri^.v[0][pa+1] := ri^.v[1][pb+1];
	 pa := pa + ri^.width div 2;
	 pb := pb + ri^.width div 2;
        end;

end;

procedure apply_motion_8x8(var ri:Proq_info; x, y: integer; mv,mean_x,mean_y : byte);
var
mx,my,i : integer;
pa,pb : integer;
begin

        mx := x + 8 - (mv shr 4) - mean_x;
	my := y + 8 - (mv and $f) - mean_y;

	pa := (y * ri^.width) + x; //ri^.y[0]
	pb := (my * ri^.width) + mx; //ri^.y[1]
        for x := 0 to 7 do
	begin
         ri^.y[0][pa+0] := ri^.y[1][pb+0];
         ri^.y[0][pa+1] := ri^.y[1][pb+1];
         ri^.y[0][pa+2] := ri^.y[1][pb+2];
         ri^.y[0][pa+3] := ri^.y[1][pb+3];
         ri^.y[0][pa+4] := ri^.y[1][pb+4];
         ri^.y[0][pa+5] := ri^.y[1][pb+5];
         ri^.y[0][pa+6] := ri^.y[1][pb+6];
         ri^.y[0][pa+7] := ri^.y[1][pb+7];
	 pa := pa + ri^.width;
	 pb := pb + ri^.width;
	end;

        pa := (y div 2) * (ri^.width div 2) + x div 2;
	pb := (my div 2) * (ri^.width div 2) + (mx + 1) div 2;
        for x := 0 to 3 do
	begin
         ri^.u[0][pa+0] := ri^.u[1][pb+0];
         ri^.u[0][pa+1] := ri^.u[1][pb+1];
         ri^.u[0][pa+2] := ri^.u[1][pb+2];
         ri^.u[0][pa+3] := ri^.u[1][pb+3];
	 pa := pa + ri^.width div 2;
	 pb := pb + ri^.width div 2;
	end;

        pa := (y div 2) * (ri^.width div 2) + x div 2;
	pb := (my div 2) * (ri^.width div 2) + (mx + 1) div 2;
        for x := 0 to 3 do
	begin
         ri^.v[0][pa+0] := ri^.v[1][pb+0];
         ri^.v[0][pa+1] := ri^.v[1][pb+1];
         ri^.v[0][pa+2] := ri^.v[1][pb+2];
         ri^.v[0][pa+3] := ri^.v[1][pb+3];
	 pa := pa + ri^.width div 2;
	 pb := pb + ri^.width div 2;
	end;


end;

Function roq_open(fname : pchar) : Proq_info;
var
F : File;
RI : Proq_info;
i : Integer;
begin

if not fileexists(fname) then
begin
messagebox(0,pchar('Error opening ' + fname),'Error Opening File',0);
Result := nil;
exit;
end;

new(ri);
AssignFile(ri^.FP,fname);
Reset(ri^.FP,1);

//ri^.FP := F;

if (roq_parse_file(ri) = 1) then
begin
Result := nil;
exit;
end;

ri^.stream_length := (ri^.num_frames * 1000) div 30;

for i := 0 to 127 do
begin
ri^.snd_sqr_arr[i]       := i * i;
ri^.snd_sqr_arr[i + 128] := -(i * i);
end;

for i := 0 to 1 do
begin
SetLength(ri^.y[i],ri^.width * ri^.height);
SetLength(ri^.u[i],ri^.width * ri^.height div 4);
SetLength(ri^.v[i],ri^.width * ri^.height div 4);
end;

ri^.buf_size := ri^.buf_size * 2;
SetLength(ri^.buf,ri^.buf_size);

ri^.audio_buf_size := 0;
ri^.audio          := nil;
ri^.frame_num      := 0;
ri^.aud_pos        := ri^.roq_start;
ri^.vid_pos        := ri^.roq_start;

Result := Ri;

end;

Procedure roq_close(var RI : Proq_info);
var
I : integer;
begin

if ri = nil then exit;
CloseFile(ri^.fp);

for i := 0 to 1 do
begin
Setlength(ri^.y[i],0);
Setlength(ri^.u[i],0);
Setlength(ri^.v[i],0);
end;
Setlength(ri^.buf,0);

ri := nil;
end;

Function roq_read_frame(ri : Proq_info) : integer;
var
fp : PFile;
chunk_id,chunk_arg : cardinal;
chunk_size : longword;
i, j, k, nv1, nv2, vqflg, vqflg_pos, vqid, bpos, xpos, ypos, xp, yp, x, y : integer;
tp, buf : array of Byte;
frame_stats : array [0..1] of array [0..3] of integer;
qcell : Proq_qcell;
begin
fp := @ri^.FP;

chunk_id := 0;
chunk_arg := 0;
chunk_size := 0;

i := 0;
j := 0;
k := 0;
nv1 := 0;
nv2 := 0;
vqflg := 0;
vqflg_pos := -1;

Setlength(tp,0);
Setlength(buf,0);

//frame_stats[2][4] = {{0},{0}};

fseek(fp, ri^.vid_pos, SEEK_SET);

while not feof(fp) do
begin

                chunk_id   := get_word(fp);
		chunk_size := get_long(fp);
		chunk_arg  := get_word(fp);
		if(feof(fp)) then break;
		if(chunk_id = RoQ_QUAD_VQ) then break;
		if(chunk_id = RoQ_QUAD_CODEBOOK) then
		begin
                nv1 := chunk_arg shr 8;
                nv2 := chunk_arg and $ff;

                if nv1 = 0 then nv1 := 256;
                if (nv1 = 0) and (nv1 * 6 < chunk_size) then nv2 := 256;

                        for i := 0 to nv1-1 do
                        begin
                                ri^.cells[i].y0 := fgetc(fp);
				ri^.cells[i].y1 := fgetc(fp);
				ri^.cells[i].y2 := fgetc(fp);
				ri^.cells[i].y3 := fgetc(fp);
				ri^.cells[i].u  := fgetc(fp);
				ri^.cells[i].v  := fgetc(fp);
                        end;

                        for i := 0 to nv2-1 do
                        for j := 0 to 3 do
			ri^.qcells[i].idx[j] := fgetc(fp);
			end
		else fseek(fp, chunk_size, SEEK_CUR);
end;

if (chunk_id <> RoQ_QUAD_VQ) then
begin
ri^.vid_pos := ftell(fp);
Result := 0;
exit;
end;

inc(ri^.frame_num);

if(ri^.buf_size < chunk_size) then
begin
ri^.buf_size := ri^.buf_size * 2;
setlength(ri^.buf,ri^.buf_size);
end;
BlockRead(fp^,ri^.Buf,chunk_size);//fread(ri^.buf, chunk_size, 1, fp);

buf := @ri^.buf;

bpos := 0;
xpos := 0;
ypos := 0;

	while (bpos < chunk_size) do
	begin
                yp := ypos;
                while (yp < ypos + 16) do
                begin
                xp := xpos;
                        while (xp < xpos + 16) do
                        begin
				if(vqflg_pos < 0) then
					begin
					vqflg := buf[bpos];
                                        inc(bpos);
                                        vqflg := vqflg or (buf[bpos] shl 8);
                                        inc(bpos);
					vqflg_pos := 7;
					end;
				vqid := (vqflg shr (vqflg_pos * 2)) and $3;
				inc(frame_stats[0][vqid]);
                                dec(vqflg_pos);

                                case vqid of
					RoQ_ID_MOT: break;
					RoQ_ID_FCC:
                                                begin
						apply_motion_8x8(ri, xp, yp, buf[bpos], chunk_arg shr 8, chunk_arg and $ff);
                                                inc(bpos);
						break;
                                                end;
					RoQ_ID_SLD:
                                                begin
                                                // ERROR
                                                {
						qcell = ri^.qcells + buf[bpos];
                                                inc(bpos);
						apply_vector_4x4(ri, xp, yp, ri^.cells + qcell^.idx[0]);
						apply_vector_4x4(ri, xp+4, yp, ri^.cells + qcell^.idx[1]);
						apply_vector_4x4(ri, xp, yp+4, ri^.cells + qcell^.idx[2]);
						apply_vector_4x4(ri, xp+4, yp+4, ri^.cells + qcell^.idx[3]);
                                                }
						break;
                                                end;
					RoQ_ID_CCC:
                                        begin
                                                for k := 0 to 3 do
                                                begin
							x := xp;
                                                        y := yp;
							if (k and $01) = 1 then x := x + 4;
							if (k and $02) = 1 then y := y + 4;

							if(vqflg_pos < 0) then
								begin
								vqflg := buf[bpos];
                                                                inc(bpos);
                                                                vqflg := vqflg or (buf[bpos] shl 8);
                                                                inc(bpos);
								vqflg_pos := 7;
								end;
							vqid := (vqflg shr (vqflg_pos * 2)) and $3;
                                                        inc(frame_stats[1][vqid]);
							dec(vqflg_pos);
                                                        case vqid of
								RoQ_ID_MOT: break;
								RoQ_ID_FCC:
                                                                        begin
									apply_motion_4x4(ri, x, y, buf[bpos], chunk_arg shr 8, chunk_arg and $ff);
                                                                        inc(bpos);
									break;
                                                                        end;
								RoQ_ID_SLD:
                                                                        begin
                                                                        //ERROR
									{
                                                                        qcell = ri^.qcells + buf[bpos];
                                                                        inc(bpos);
									apply_vector_2x2(ri, x, y, ri^.cells + qcell^.idx[0]);
									apply_vector_2x2(ri, x+2, y, ri^.cells + qcell^.idx[1]);
									apply_vector_2x2(ri, x, y+2, ri^.cells + qcell^.idx[2]);
									apply_vector_2x2(ri, x+2, y+2, ri^.cells + qcell^.idx[3]);
                                                                        }
									break;
                                                                        end;
								RoQ_ID_CCC:
                                                                        begin
                                                                        // Error
                                                                        {
									apply_vector_2x2(ri, x, y, ri^.cells + buf[bpos]);
									apply_vector_2x2(ri, x+2, y, ri^.cells + buf[bpos+1]);
									apply_vector_2x2(ri, x, y+2, ri^.cells + buf[bpos+2]);
									apply_vector_2x2(ri, x+2, y+2, ri^.cells + buf[bpos+3]);
									bpos := bpos + 4;
                                                                        }
									break;
                                                                        end;
								end;
							end;
						break;
                                                end;
					else
                                        messagebox(0,pchar('Unknown vq code: ' + inttostr(vqid)),'Error',0);

					end;
                              xp := xp + 8;
                              end;
                         yp := yp + 8;
                         end;

		xpos := xpos + 16;
		if (xpos >= ri^.width) then
			begin
			xpos := xpos - ri^.width;
			ypos := ypos + 16;
			end;
		if(ypos >= ri^.height) then break;
	
end;




	ri^.vid_pos := ftell(fp);

	if(ri^.frame_num = 1) then
        begin
        //ERROR
		{
		memcpy(ri^.y[1], ri^.y[0], ri^.width * ri^.height);
		memcpy(ri^.u[1], ri^.u[0], (ri^.width * ri^.height)/4);
		memcpy(ri^.v[1], ri^.v[0], (ri^.width * ri^.height)/4);
		}
        end
	else
        begin
        //ERROR
		{
		tp = ri^.y[0]; ri^.y[0] = ri^.y[1]; ri^.y[1] = tp;
		tp = ri^.u[0]; ri^.u[0] = ri^.u[1]; ri^.u[1] = tp;
		tp = ri^.v[0]; ri^.v[0] = ri^.v[1]; ri^.v[1] = tp;
		}
        end;

        Result := 1;
end;

//Todo
{
int roq_read_frame(roq_info *ri) // Needs todo procedures above made
}
//

Function roq_read_audio(var Ri : Proq_info) : integer;
var
FP : PFile;
chunk_id,chunk_arg : integer;
chunk_size : longint;
i,snd_left,snd_right : integer;
begin
FP         := @ri^.FP;
chunk_id   := 0;
chunk_arg  := 0;
chunk_size := 0;

Fseek(FP,ri^.aud_pos,SEEK_SET);
ri^.audio_size := 0;

while not feof(FP) do
begin
chunk_id   := get_word(fp);
chunk_size := get_long(fp);
chunk_arg  := get_word(fp);
if feof(fp) or (chunk_id = RoQ_SOUND_MONO) or (chunk_id = RoQ_SOUND_STEREO) then break;
fseek(fp, chunk_size, SEEK_CUR);
end;

if (ri^.audio_buf_size < chunk_size * 2) then
begin
setlength(ri^.audio,0);
ri^.audio_buf_size := chunk_size * 3;
setlength(ri^.audio,ri^.audio_buf_size);
end;

if(chunk_id = RoQ_SOUND_MONO) then
begin
ri^.audio_size := chunk_size * 2;
snd_left       := chunk_arg;
for i := 0 to chunk_size-1 do
begin
 snd_left             := snd_left + ri^.snd_sqr_arr[fgetc(fp)];
 ri^.audio[i * 2]     := snd_left and $ff;
 ri^.audio[i * 2 + 1] := (snd_left and $ff00) shr 8;
end;
ri^.aud_pos := ftell(fp);
Result := 1;
exit;
end;

if(chunk_id = RoQ_SOUND_STEREO) then
begin
ri^.audio_size := chunk_size * 2;
snd_left       := (chunk_arg and $FF00);
snd_right      := (chunk_arg and $FF) shl 8;
i := 0;
while i < chunk_size do
begin
 snd_left             := snd_left + ri^.snd_sqr_arr[fgetc(fp)];
 snd_right            := snd_right + ri^.snd_sqr_arr[fgetc(fp)];
 ri^.audio[i * 2]     := snd_left and $ff;
 ri^.audio[i * 2 + 1] := (snd_left and $ff00) shr 8;
 ri^.audio[i * 2 + 2] := snd_right and $ff;
 ri^.audio[i * 2 + 3] := (snd_right and $ff00) shr 8;
inc(i,2);
end;
ri^.aud_pos := ftell(fp);
Result := 1;
exit;
end;

ri^.aud_pos := ftell(fp);
Result := 0;
end;


end.
